package net.meltingwax.anonircd;

import java.security.*;
import java.util.*;
import java.util.Map.Entry;
import net.meltingwax.anonircd.conf.*;


/**
 * An account on the IRC daemon, with his/her privileges on a set of channels.
 */
public class AuthedAccount {
    private String username;
    private String passHash;
    private boolean isSysOp;
    private Map<String, ChannelPrivilege> channelPrivileges;

    public AuthedAccount(String username, String passHash, boolean isSysOp) {
        this.username = username;
        this.passHash = passHash;
        this.isSysOp = isSysOp;
        this.channelPrivileges = new HashMap<String, ChannelPrivilege>();

        assert ConfigUtils.canHashPasswords();
    }

    /**
     * Returns whether this user is a sysop.
     */
    public boolean getIsSysOp() {
        return isSysOp;
    }

    /**
     * Gets the privilege this account has on a given channel.
     * @param channelName (The name of the channel, with preceding #).
     */
    public ChannelPrivilege getChannelPrivilege(String channelName) {
        ChannelPrivilege privilege;

        if (channelPrivileges.containsKey(channelName)) {
            privilege = channelPrivileges.get(channelName);
        } else {
            privilege = ChannelPrivilege.NONE;
        }

        return privilege;
    }

    /**
     * Gets the password hash
     */
    public String getPasswordHash() {
        return passHash;
    }

    /**
     * Returns an array of channel privilege names.
     */
    public static String[] getChannelPrivilegeNames() {
        ChannelPrivilege[] allPrivileges = ChannelPrivilege.values();
        String[] privilegeNames = new String[allPrivileges.length];

        for (int i=0; i < allPrivileges.length; i++) {
            privilegeNames[i] = allPrivileges[i].name().toUpperCase();
        }

        return privilegeNames;
    }

    /**
     * Gets the username of this account.
     */
    public String getUsername() {
        return username;
    }

    /**
     * Gets the channel the account has elevated privileges on.
     * @return
     */
    public Set<String> getChannelsWithElevatedPrivileges() {
        Set<String> channels = new HashSet<String>();

        for (Entry<String, ChannelPrivilege> entry : channelPrivileges.entrySet()) {
            if (entry.getValue() != ChannelPrivilege.NONE) {
                channels.add(entry.getKey());
            }
        }

        return channels;
    }

    /**
     * Sets a privilege for this user on a given channel.
     */
    public void setChannelPriviledge(String channelName, ChannelPrivilege privilege) {
        synchronized (channelPrivileges) {
            channelPrivileges.put(channelName, privilege);
        }
    }

    /**
     * Sets whether the account is a SysOp (System Operator).
     */
    public void setIsSysOp(boolean isSysOp) {
        this.isSysOp = isSysOp;
    }

    /**
     * Sets the password for this user. It is only accessible in the future
     * as a password hash (@see getPasswordHash).
     * @return
     */
    public void setPassword(String password) {
        try {
            passHash = ConfigUtils.hashPassword(password);
        } catch (NoSuchAlgorithmException e) {
        }
    }
}
