package net.meltingwax.anonircd;

import net.meltingwax.anonircd.i.*;
import net.meltingwax.anonircd.conf.ConfigUtils;
import net.meltingwax.anonircd.excepts.*;

import java.security.NoSuchAlgorithmException;
import java.util.*;


public class Channel {

    /**
     * A message in the channel (struct-like).
     */
    public class ChannelMessage {
        public int num;
        public ChannelParticipant participant;
        public String message;

        public ChannelMessage(int num, ChannelParticipant participant, String message) {
            this.num = num;
            this.participant = participant;
            this.message = message;
        }
    }

    private ConfigProvider configProvider;
    private String name;
    private String topic;
    private Set<ChannelParticipant> participants;
    private LinkedList<ChannelMessage> recentMessages;
    private int recentMessagesCounter;

    public Channel(ConfigProvider configProvider, String name) throws IllegalChannelNameException {
        if (! isValidChannelName(name)) {
            throw new IllegalChannelNameException(name);
        }

        String storedTopic = configProvider.getChannelTopic(name);

        this.configProvider = configProvider;
        this.name = name;
        this.topic = (storedTopic == null) ? "" : storedTopic;
        this.participants = new HashSet<ChannelParticipant>();
        this.recentMessages = new LinkedList<ChannelMessage>();
        this.recentMessagesCounter = 1;
    }

    /**
     * Determines whether a non-null string is a valid channel name.
     * @param name The channel name to test.
     * @return true if given arg is an acceptable channel name.
     */
    public static boolean isValidChannelName(String name) {
        return name.startsWith("#") && name.indexOf(' ') == -1;
    }

    /**
     * Adds a new client to this channel.
     * @param client The ChannelParticipant to add.
     * @throws IllegalArgumentException Already on channel.
     * @throws ClientIsBannedException Rejected from joining channel due to a ban.
     */
    public void addClient(ChannelParticipant client) throws ClientIsBannedException {
        if (participants.contains(client)) {
            throw new IllegalArgumentException();
        }

        if (configProvider.checkForChannelBan(name, client.getIPAddress())) {
            throw new ClientIsBannedException(client, this);
        }

        // Temp code (TODO: remove & implement subnet IP channing)
        if (client.getIPAddress().startsWith("63.153.")) {
            throw new ClientIsBannedException(client, this);
        }
        // end

        participants.add(client);

        // Call callback for this event on each participant
        for (ChannelParticipant participant : participants) {
            if (participant != client) {
                participant.onJoinedChannelNewChannelParticipant(this);
            }
        }
    }

    /**
     * Bans and kicks sender of a given message num. Applies to all
     * connections from that IP address.
     * @param num The message number
     * @param range Specify this to true to ban on the XX.YY.*.* range
     * @throws ChannelParticipantNotFoundException num either has not
     * occurred yet or is too old.
     */
    public void ban(int num, boolean range) throws ChannelParticipantNotFoundException {
        // Search recent message list for participant to ban
        ChannelParticipant toBan = null;

        for (ChannelMessage recentMessage : recentMessages) {
            if (recentMessage.num == num) {
                toBan = recentMessage.participant;
                break;
            }
        }

        if (toBan == null) {
            throw new ChannelParticipantNotFoundException();
        }

        // Kick and ban the participant by IP.
        String targetIP = toBan.getIPAddress();

        if (range) {
            String[] toks = targetIP.split("\\.");
            toks[3] = toks[2] = "*";
            targetIP = Utils.join(".", toks);
        }

        configProvider.addChannelBan(name, targetIP);
        kick(toBan.getIPAddress());
    }

    /**
     * Kick all connections from a given IP address.
     */
    public void kick(String ipAddr) {
        // Build list of participants to kick
        List<ChannelParticipant> toKick = new LinkedList<ChannelParticipant>();

        for (ChannelParticipant participant : participants) {
            if (participant.getIPAddress().equals(ipAddr)) {
                toKick.add(participant);
            }
        }

        // Call kick handler and remove them from the channel
        for (ChannelParticipant participant : toKick) {
            participant.onKick(this);
            removeClient(participant);
        }
    }

    /**
     * Kick connection from recent message matching num.
     * @param num The message number.
     * @throws ChannelParticipantNotFoundException num either has not
     * occurred yet or is too old
     */
    public void kick(int num) throws ChannelParticipantNotFoundException {
        // Determine the participant to kick. num=-1 is the last user, if one exists.
        ChannelParticipant participant = null;


        // the recent messages queue to find the target participant
        for (ChannelMessage recentMessage : recentMessages) {
            if (recentMessage.num == num) {
                participant = recentMessage.participant;
                break;
            }
        }

        if (participant == null) {
            throw new ChannelParticipantNotFoundException();
        }

        // Call kick handler and remove them from the channel
        participant.onKick(this);
        removeClient(participant);
    }

    /**
     * Remove a client from the channel.
     * @param client
     */
    public void removeClient(ChannelParticipant client) {
        if (! participants.contains(client)) {
            throw new IllegalArgumentException();
        }

        synchronized (participants) {
            participants.remove(client);
        }

        for (ChannelParticipant participant : participants) {
            participant.onJoinedChannelParticipantPart(this);
        }
    }

    public void sendMessage(String message, ChannelParticipant from) throws MessageTooLongException {
        if (message.length() > configProvider.getMaxMessageLength()) {
            throw new MessageTooLongException(message, configProvider.getMaxMessageLength());
        }

        if (! participants.contains(from)) {
            throw new IllegalArgumentException();
        }

        
      	/**
    	 * Here we determine the nick that the message is labeled to have come from.
    	 * If we are a special channel, we attach a custom identifier to the source
    	 * nick. The custom identifier is set to change every hour on the hour, and
    	 * is unique per connection and channel.
    	 */
        String fromNick = from.getOutsideNick(this);
                
        for (ChannelParticipant participant : participants) {
            if (participant != from) {
                participant.onJoinedChannelMessage(this, message, fromNick);
            }
        }

        // Update recent messages
	addRecentMessage(from, message);
    }

    public String getAnonParticipantNick(int numOtherUsers) {
        return numOtherUsers + "_other_users";
    }

    /**
     * Returns the total number of participants.
     */
    public int getNumParticipants() {
        int num = 0;

        for (ChannelParticipant participant : participants) {
            if (participant.isParticipantAlive()) {
                num++;
            }
        }

        return num;
    }

    public String getName() {
        return name;
    }

    public List<ChannelMessage> getRecentMessages() {
        return new ArrayList<ChannelMessage>(recentMessages);
    }

    public String getTopic() {
        return topic;
    }

    public void setTopic(String topic, ChannelParticipant source) throws TopicTooLongException {
        if (topic.length() > configProvider.getMaxTopicLength()) {
            throw new TopicTooLongException(topic, configProvider.getMaxTopicLength());
        }

	// The order of operations here is to store the instance variables,
	// save to configuration store, and then notify each client of the
	// changes to the channel's topic.
	this.topic = topic;
	addRecentMessage(source, "/topic " + topic);

	configProvider.saveChannelTopic(name, topic);

        for (ChannelParticipant participant : participants) {
            try {
                participant.onJoinedChannelNewTopic(this);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * Updates recent messages list. Trims old messages out of list if
     * necessary to add the new one and not make it too long.
     *
     * @param source   The channel participant the message to add originated
     *                 from.
     * @param message  The message to add to the list.
     */
    private void addRecentMessage(ChannelParticipant source, String message) {
        synchronized (recentMessages) {
            recentMessages.addLast(new ChannelMessage(recentMessagesCounter, source, message));
            recentMessagesCounter++;
	    
            if (recentMessages.size() > configProvider.getMaxChannelRecentMessages()) {
                recentMessages.removeFirst();
            }
        }
    }
    
    /**
     * Returns whether this is a special ID channel.
     */
    public boolean isSpecialIDChannel() {
    	return name.endsWith("!");
    }
}
