package net.meltingwax.anonircd;

import net.meltingwax.anonircd.i.*;
import net.meltingwax.anonircd.bots.i.*;
import net.meltingwax.anonircd.conf.ConfigUtils;
import net.meltingwax.anonircd.excepts.*;

import java.net.*;
import java.io.*;
import java.security.NoSuchAlgorithmException;
import java.util.*;


public class ConnectionHandler extends Thread implements BotClient, ChannelParticipant {
    private Socket clientSocket;
    private BotProvider botProvider;
    private ConfigProvider configProvider;
    private ChannelProvider channelProvider;
    private String ipAddr;
    private BufferedReader in;
    private PrintWriter out;
    private String nick;
    private Set<Channel> activeChannels;  // Channels the user is in.

    /**
     * An exception for handleCOMMAND() methods to throw in order to trigger
     * a disconnect and termination of the main loop.
     */
    @SuppressWarnings("serial")
    private class HandlerTriggerDisconnect extends Exception { }

    /**
     * Instantiate the ConnectionHandler.
     * @param clientSocket The accepted socket for the connection
     * @throws IOException There was an error getting streams for the socket
     */
    public ConnectionHandler(Socket clientSocket, ConfigProvider configProvider,
    		ChannelProvider channelProvider, BotProvider botProvider) throws IOException {
        this.clientSocket = clientSocket;
        this.botProvider = botProvider;
        this.channelProvider = channelProvider;
        this.configProvider = configProvider;

        nick = null;

        // This may throw an IO error
        in = new BufferedReader(new InputStreamReader(clientSocket.getInputStream()));
        out = new PrintWriter(clientSocket.getOutputStream(), true);
        // end

        activeChannels = new HashSet<Channel>();
    }

    /**
     * Disconnects the client, performing cleanup and properly /part'ing
     * from joined channels.
     */
    public void disconnect() {
        try {
            clientSocket.close();
        } catch (IOException e) {
            // already closed
        }

        System.out.println(ipAddr + "\t###\t" + "Disconnected");


        // Leave parted channels
        Set<Channel> channelsToLeave;

        synchronized (activeChannels) {
            channelsToLeave = new HashSet<Channel>(activeChannels);
        }

        for (Channel activeChannel : channelsToLeave) {
            try {
                activeChannel.removeClient(this);
            } catch (IllegalArgumentException e) {
                // guess we're already gone.
            }
        }
    }

    /**
     * Function for receiving a line from the socket. This blocks until a line is available. If there was
     * an error or if connection was lost, returns Null.
     * @return Line
     */
    private String recvLine() {
        String line;

        try {
            line = in.readLine();
        } catch (IOException e) {
            line = null;
        }

        if (line != null) {
            System.out.println(ipAddr + "\t<--\t" + line);
        }

        return line;
    }

    /**
     * Function for sending a line to the socket. Accepts a list of words which
     * will be joined by spaces.
     *
     * @param words
     */
    private void sendLine(String... words) {
        if (words.length == 0) {
            throw new IllegalArgumentException();
        }

        // Leaving out the :blah at the beginning breaks xchat.
        String line = (words[0].startsWith(":")) ? "" : " :null";

        for (String command : words) {
            line += " " + command;
        }

        line = line.substring(1);

        System.out.println(ipAddr + "\t-->\t" + line);

	synchronized (out) {
	    out.write(line + IRCConsts.EOL);
	    out.flush();
	}
    }

    public void onSystemNotice(String message) {
	Set<Channel> weakCopy;
	
	synchronized (activeChannels) {
	    weakCopy = new HashSet<Channel>(activeChannels);
	}
	
	for (Channel channel : weakCopy) {
	    sendLine(":SystemNotice", "NOTICE", channel.getName(), ":" + message);
	}
	
	sendLine(":SystemNotice", "NOTICE", "SYS", ":" + message);
    }
    
    public void run() {
        // Do the connection sequence. This is over when we have seen both
        // NICK and USER lines. If there is a PASS line it comes before both
        // of those lines.
        //
        // If the connection is coming from a trusted web client, we allow the
        // client to use the USER line to override the IP address associated
        // with this connection.
        ipAddr = clientSocket.getInetAddress().toString().substring(1);
        boolean isTrustedWebClient = configProvider.checkForTrustedWebClient(ipAddr);

        boolean hadNick = false;
        boolean hadUser = false;

        while (! (hadNick && hadUser)) {
            String line = null;

            line = recvLine();

            if (line == null) {
                disconnect();
                return;
            } else if (line.startsWith("NICK ")) {
                hadNick = true;

                String[] toks = line.split(" ");
                if (toks.length > 1) {
                    nick = toks[1];
                } else {
                    disconnect();
                    return;
                }
            } else if (line.startsWith("USER ")) {
                hadUser = true;

                // Parse the IP address out of the USER line
                // and override our associated IP Address.
                String[] words = line.split(" ");

                if (words.length >= 5) {
                    String[] toks = words[4].split("/");

                    if (toks.length >= 2) {
                        ipAddr = toks[1];
                    }
                }
            }
        }

        assert nick != null;
        
        System.out.println(ipAddr + "\t###\t" + "Connection established."
                           + (isTrustedWebClient ? " [from WebClient]" : ""));
        
        // Send initial connection messages.
        sendLine("NOTICE AUTH :Welcome Anonymous");
        sendLine(IRCConsts.WELCOME, nick, "Welcome to AnonIRCd");
        sendLine(IRCConsts.YOURHOST, nick, ":Your host is invisible");
        sendLine(IRCConsts.MYINFO, nick, "");
        sendLine(IRCConsts.FEATURELIST, nick, ":All messages will be sent as Anonymous");

        handleMOTD();

        // Loop on messages
        while (true){

            // Receive line
            String line = recvLine();

            if (line == null) {
                disconnect();
                return;
            }

            if (line.isEmpty()) {
                continue;
            }

            // Convert line to list of tokens. Examples:
            // "a b c"         => ArrayList({"a", "b", "c"})
            // "a b :def ghf"  => ArrayList({"a", "b", "def ghf"})
            List<String> toks = new ArrayList<String>();
            String[] words = line.split(" ");
            boolean stopFlag = false;

            for (int i = 0; i < words.length && ! stopFlag; i++) {
                if (words[i].startsWith(":")) {
                    String tok = words[i].substring(1);

                    for (int j = i + 1; j < words.length; j++) {
                        tok += " " + words[j];
                    }

                    toks.add(tok);
                    stopFlag = true;
                } else {
                    toks.add(words[i]);
                }
            }

            List<String> argsList = toks.subList(1, toks.size());
            String[] args = new String[argsList.size()];

            int i = 0;
            while (! argsList.isEmpty()) {
                args[i] = argsList.remove(0);
                i++;
            }

            // Call appropriate handler for this command.
            String command = toks.get(0).toUpperCase();

            try {
                if (command.equals("JOIN")) {
                    handleJoin(args);
                } else if (command.equals("LIST")) {
                    handleList();
                } else if (command.equals("PING")) {
                    handlePing(args);
                } else if (command.equals("PART")) {
                    handlePart(args);
                } else if (command.equals("PRIVMSG")) {
                    handlePrivMsg(args);
                } else if (command.equals("MOTD")) {
                    handleMOTD();
                } else if (command.equals("TOPIC")) {
                    handleTopic(args);
                } else if (command.equals("QUIT")) {
                    handleQuit();
                } else {
                    System.out.println("Did not know how to handle command " + command);
                }
            } catch (HandlerTriggerDisconnect e) {
                disconnect();
                return;
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * Handler for receiving a JOIN command.
     */
    public void handleJoin(String[] args) {
        if (args.length != 1) {
            return;
        }

        String[] channelNames = args[0].split(",");

        for (String channelName : channelNames) {
            // Get the channel instance
            Channel channel;

            try {
                channel = channelProvider.getChannel(channelName);
            } catch (IllegalChannelNameException e) {
                continue;  // Skip this channel
            }

            // Try joining the channel
            try {
                channel.addClient(this);
            } catch (ClientIsBannedException e) {
                sendLine(IRCConsts.BANNEDFROMCHAN, nick, channelName, ":Cannot join channel - you are banned.");
                continue;
            } catch (IllegalArgumentException e) {
                continue;
            }

            synchronized (activeChannels) {
                activeChannels.add(channel);
            }
            
   
            sendLine(":" + Utils.makeAnonUserString(nick), "JOIN", ":" + channel.getName());
            sendLine(IRCConsts.CURRENTTOPIC, nick, channel.getName(), ":" + channel.getTopic());
            sendLine(IRCConsts.TOPICINFO, nick, channel.getName(), "Anonymous", "0");
            sendLine(IRCConsts.NAMREPLY, nick, "=", channel.getName(),
                     ":" + nick + " " + channel.getAnonParticipantNick(channel.getNumParticipants() - 1));
            sendLine(IRCConsts.ENDOFNAMES, nick, channel.getName(), ":End of /NAMES list");
        }
    }

    /**
     * Handler for receiving the LIST comamnd.
     */
    public void handleList() {
        // Build list of non-empty channels sorted by number of participants, descending.
        Map<String, Channel> channelMap = channelProvider.getAllChannels();
        ArrayList<Channel> channelList = new ArrayList<Channel>();

        for (String channelName : channelMap.keySet()) {
            Channel channel = channelMap.get(channelName);
            if (channel.getNumParticipants() > 0) {
                channelList.add(channel);
            }
        }

        Collections.sort(channelList, new Comparator<Channel>() {
                public int compare(Channel a, Channel b) {
                    int n_a = a.getNumParticipants();
                    int n_b = b.getNumParticipants();
                    return n_b - n_a;  // descending
                }
            });

        // Send list block
        sendLine(IRCConsts.LISTSTART, nick, "Channel", ":Num Users");

        for (Channel channel : channelList) {
            sendLine(IRCConsts.LIST, nick, channel.getName(), channel.getNumParticipants() + "", ":[+?]");
        }

        sendLine(IRCConsts.LISTEND, nick, ":End of channel list");
    }

    /**
     * Handler for receiving a MOTD command.
     */
    public void handleMOTD() {
        sendLine(IRCConsts.MOTDSTART, nick, ":Message of the day -");

        for (String motdLine : configProvider.getMOTD().split("\n")) {
            sendLine(IRCConsts.MOTD, nick, ":-  " + motdLine);
        }

        sendLine(IRCConsts.ENDOFMOTD, nick, ":End of /MOTD command");
    }

    /**
     * Handler for receiving a PART command.
     */
    public void handlePart(String[] args) {
        if (args.length != 1) {
            return;
        }

        Channel channel;

        try {
            channel = channelProvider.getChannel(args[0]);
        } catch (IllegalChannelNameException e) {
            return;  // Turn this into a no-op
        }

        channel.removeClient(this);

        synchronized (activeChannels) {
            activeChannels.remove(channel);
        }
    }

    /**
     * Handler for receiving a PRIVMSG command.
     */
    public void handlePrivMsg(String[] args) {
        if (args.length != 2) {
            return;
        }

        String target = args[0];
        String message = args[1];

        if (target.startsWith("#")) {
            Channel channel;

            try {
                channel = channelProvider.getChannel(target);
            } catch (IllegalChannelNameException e) {
                return;  // Turn this into a no-op
            }

            try {
                channel.sendMessage(message, this);
            } catch (MessageTooLongException e) {
                return;  // Turn this into a no-op
            }
        } else {
            try {
                Bot bot = botProvider.getBot(target);
                bot.onMessage(message, this);
            } catch (BotNotFoundException e) {
                sendLine(IRCConsts.NOSUCHNICK, nick, target, ":No such nick");
            }
        }
    }

    /**
     * Handler for receiving a TOPIC command.
     */
    public void handleTopic(String[] args) {
        if (args.length != 2) {
            return;
        }

        String channelName = args[0];
        String topic = args[1];

        Channel channel;

        try {
            channel = channelProvider.getChannel(channelName);
        } catch (IllegalChannelNameException e) {
            return;  // Turn this into a no-op
        }

        try {
            channel.setTopic(topic, this);
        } catch (TopicTooLongException e){
            return;  // Turn this into a no-op
        }
    }

    /**
     * Handler for receiving a PING command.
     */
    public void handlePing(String[] args) {
        if (args.length != 1) {
            return;
        }

        sendLine("PONG", "null", args[0]);
    }

    /**
     * Handler for receiving a QUIT command.
     * @throws HandlerTriggerDisconnect
     */
    public void handleQuit() throws HandlerTriggerDisconnect {
        throw new HandlerTriggerDisconnect();
    }

    /**
     * Callback for another participant leaving one of our channels.
     * @param channel Channel for which a user is leaving.
     */
    public void onJoinedChannelParticipantPart(Channel channel) {
        int n = channel.getNumParticipants();

        sendLine(":" + Utils.makeAnonUserString(channel.getAnonParticipantNick(n)),
                 "NICK",
                 ":" + channel.getAnonParticipantNick(n - 1));
    }

    /**
     * Callback for when a user in one of our channels said something. That user is
     * always someone else.
     * @param channel Channel for which there is a new message.
     * @param message Contents of the new message.
     */
    public void onJoinedChannelMessage(Channel channel, String message, String fromNick) {
        sendLine(":" + Utils.makeAnonUserString(fromNick), "PRIVMSG", channel.getName(), ":" + message);
    }

    /**
     * Callback for when a new user has joined one of our channels. That user
     * is always someone else.
     * @param channel The channel for which there is a new member.
     */
    public void onJoinedChannelNewChannelParticipant(Channel channel) {
        int n = channel.getNumParticipants();

        sendLine(":" + Utils.makeAnonUserString(channel.getAnonParticipantNick(n - 2)),
                 "NICK",
                 ":" + channel.getAnonParticipantNick(n - 1));
    }

    /**
     * Callback for when the topic has changed in one of our channels.
     * @param channel The channel whose topic has changed.
     */
    public void onJoinedChannelNewTopic(Channel channel) {
        sendLine(":" + Utils.makeAnonUserString("Anonymous"), "TOPIC", channel.getName(), ":" + channel.getTopic());
    }

    /**
     * Callback for being kicked from a channel.
 	 * @param channel The channel the user is being kicked from.
     */
    public void onKick(Channel channel) {
        sendLine(":" + Utils.makeAnonUserString("Anonymous"), "KICK", channel.getName(), nick, ":");
    }

    /**
     * Returns the IP address of the client.
     */
    public String getIPAddress() {
        return ipAddr;
    }

    /**
     * Whether or not the user should be counted as having an alive
     * connection.
     */
    public boolean isParticipantAlive() {
        return (clientSocket.isConnected()
                && (! clientSocket.isClosed())
                && (! clientSocket.isInputShutdown())
                && (! clientSocket.isOutputShutdown()));
    }

    /**
     * Callback for receiving a message from a bot.
     */
    public void recieveBotMessage(Bot from, String message) {
        sendLine(":" + Utils.makeBotUserString(from.getNick()), "PRIVMSG", nick, ":" + message);
    }

    public void recieveBotMessage(Bot from, List<String> messages) {
        for (String message : messages) {
            recieveBotMessage(from, message);
        }
    }

    /**
     * Returns the nick of this channel participant on a given channel.
     * If the channel is an ID channel, then the nick will be
     * generated according to a certain set of rules.
     */
	public String getOutsideNick(Channel channel) {
        String nick = "Anonymous";
        
    	if (channel.isSpecialIDChannel()) {
    		int hourOfDay = Calendar.getInstance().get(Calendar.HOUR_OF_DAY);
    		long fromUniqueIdentifier = getId();
    		
    		String hash;
    		try {
    			hash = ConfigUtils.hashString(hourOfDay + channel.getName() + fromUniqueIdentifier);
    		} catch (NoSuchAlgorithmException e) {
    			throw new RuntimeException(e);
    		}
    		
    		nick += "(" + hash.substring(0, 8) + ")";
    	}

    	return nick;
	}
}
