package net.meltingwax.anonircd.bots;

import net.meltingwax.anonircd.*;
import net.meltingwax.anonircd.AuthedAccount.*;
import net.meltingwax.anonircd.Channel.*;
import net.meltingwax.anonircd.bots.excepts.*;
import net.meltingwax.anonircd.bots.i.*;
import net.meltingwax.anonircd.conf.*;
import net.meltingwax.anonircd.excepts.*;
import java.security.NoSuchAlgorithmException;
import java.util.*;


public class AdminServ implements Bot {
    private Server server;
    private Map<BotClient, AuthedAccount> identifiedClients;
    private AdminServHelp help;

    public AdminServ(Server server) {
        this.server = server;
        this.identifiedClients = new HashMap<BotClient, AuthedAccount>();
        this.help = new AdminServHelp();
    }
    
    /**
     * @return The arguments of a line in a String[]. Excludes the actual 
     * command in the line.
     */
    public static String[] getLineArgs(String line) {
        String[] toks = line.trim().split(" ");
        LinkedList<String> words = new LinkedList<String>();

        for (String tok : toks) {
            if (! tok.equals("")) {
                words.add(tok);
            }
        }

        words.removeFirst();

        String[] args = new String[words.size()];

        int i = 0;
        while (! words.isEmpty()) {
            args[i] = words.removeFirst();
            i++;
        }

        return args;
    }

    public void onMessage(String message, BotClient client) {
        // Branch based on the command
        String command = message.split(" ")[0].toUpperCase();

        if (command.equals("ASSIGN")) {
            handleAssign(client, getLineArgs(message));
        } else if (command.equals("BAN")) {
            handleBan(client, getLineArgs(message));
	} else if (command.equals("BANRANGE")) {
	    handleBanRange(client, getLineArgs(message));
        } else if (command.equals("CLAIM")) {
            handleClaim(client, getLineArgs(message));
        } else if (command.equals("HELP")) {
            handleHelp(client, getLineArgs(message));
        } else if (command.equals("IDENTIFY")) {
            handleIdentify(client, getLineArgs(message));
        } else if (command.equals("KICK")) {
            handleKick(client, getLineArgs(message));
	} else if (command.equals("NOTICE")) {
	    handleNotice(client, getLineArgs(message));
	} else if (command.equals("REGISTER")) {
            handleRegister(client, getLineArgs(message));
        } else if (command.equals("REVEAL") || command.equals("LIST")) {
            handleReveal(client, getLineArgs(message));
	} else if (command.equals("UNBANIP")) {
	    handleUnbanIP(client, getLineArgs(message));
        } else {
            handleUnknown(client, command);
        }
    }

    private boolean identifiedRequiredHelper(BotClient client) {
        if (identifiedClients.containsKey(client)) {
            return true;
        } else {
            client.recieveBotMessage(this, "Your connection must be associated with an admin account to use this command. Use the IDENTIFY command (for how, type \"HELP IDENTIFY\")" );
            return false;
        }
    }

    /**
     * The format of this command is:
     *   ASSIGN #channel privilege username
     */
    public void handleAssign(BotClient client, String[] args) {
        if (! identifiedRequiredHelper(client)) {
            return;
        }

        // Reply with usage and return if number of args is not corect.
        if (args.length != 3) {
            try {
                client.recieveBotMessage(this, help.getCommandUsage("assign"));
            } catch (CommandNotFoundException e) {
                client.recieveBotMessage(this, "Invalid number of args.");
            }
            return;
        }

        // Check if client has permission to assign privileges on this channel
        String channelName = args[0];
        String targetPrivilegeName = args[1];
        String targetUsername = args[2];

        AuthedAccount account = identifiedClients.get(client);

        if (! PermissionManager.canAssign(account, channelName)) {
            String message = "You do not have permission to assign privileges on this channel.";
            client.recieveBotMessage(this, message);
            return;
        }

        // Check to see if channel name is valid
        if (! Channel.isValidChannelName(channelName)) {
            client.recieveBotMessage(this, "Invalid channel name");
        }

        // Check to see if permission is valid
        ChannelPrivilege targetPrivilege;

        try {
            targetPrivilege = ChannelPrivilege.valueOf(targetPrivilegeName.toUpperCase());
        } catch (IllegalArgumentException e) {
            String message = "Invalid privilege. Type 'ASSIGN' for a list of valid privileges";
            client.recieveBotMessage(this, message);
            return;
        }

        // Check that target account exists
        AuthedAccount targetAccount = server.config.getAuthedAccount(targetUsername);

        if (targetAccount == null) {
            client.recieveBotMessage(this, "That account does not exist.");
            return;
        }

        // Go ahead and grant new privilege to target account.
        targetAccount.setChannelPriviledge(channelName, targetPrivilege);
        server.config.saveAuthedAccount(targetAccount);

        String message = "Account granted privilege " + targetPrivilege.name().toUpperCase()
            + " on channel " + channelName;
        client.recieveBotMessage(this, message);
    }

    /**
     * The format of this command is:
     *   HELP
     *   HELP command
     */
    public void handleHelp(BotClient client, String[] args) {
        List<String> lines = new ArrayList<String>();

        if (args.length == 1) {
            try {
                lines.addAll(help.getCommandUsage(args[0]));
            } catch (CommandNotFoundException e) {
                lines.add("Command not found.");
            }
        } else {
            lines.addAll(help.getFullHelp());
        }

        client.recieveBotMessage(this, lines);
    }

    /**
     * The format of this command is:
     *   IDENTIFY username password
     */
    public void handleIdentify(BotClient client, String[] args) {
        // Halt if number of args is not correct, and refer to usage.
        if (args.length != 2) {
            try {
                client.recieveBotMessage(this, help.getCommandUsage("identify"));
            } catch (CommandNotFoundException e) {
                client.recieveBotMessage(this, "Invalid number of args.");
            }
            return;
        }

        // Check login credentials, send error message if fails.
        String username = args[0];
        String password = args[1];

        if (! server.config.checkAuthCredentials(username, password)) {
            String failMessage = "Invalid username or password. Please try again.";
            client.recieveBotMessage(this, failMessage);
        } else {
            AuthedAccount account = server.config.getAuthedAccount(username);

            synchronized (identifiedClients) {
                identifiedClients.put(client, account);
            }

            String successMessage = "Login successful.";
            client.recieveBotMessage(this, successMessage);
        }
    }

    public void handleUnbanIP(BotClient client, String[] args) {
        if (! identifiedRequiredHelper(client)) {
            return;
        }

        // Reply with usage and return if malformed
        if (args.length != 2) {
            try {
                client.recieveBotMessage(this, help.getCommandUsage("unbanip"));
            } catch (CommandNotFoundException e) {
                client.recieveBotMessage(this, "Invalid number of args.");
            }
            return;
        }

	// Parse the arguments
	String channelName = args[0];
	String ipAddr = args[1];

        // Retrieve the channel, failing on bad channel name
        Channel channel;

        try {
            channel = server.getChannel(channelName);
        } catch (IllegalChannelNameException e) {
            client.recieveBotMessage(this, "Invalid channel.");
            return;
        }

        // Require the account to have the ability to ban
        AuthedAccount account = identifiedClients.get(client);
	
        if (! PermissionManager.canBan(account, channelName)) {
            client.recieveBotMessage(this, "You do not have sufficient privileges to ban on this channel.");
            return;
        }
	
	if (server.config.checkForChannelBan(channelName, ipAddr)) {
	    server.config.removeChannelBan(channelName, ipAddr);
	    client.recieveBotMessage(this, "Ban on the specified IP address for " + channelName + " was dropped. Ban ranges may still be in effect.");
	} else {
	    client.recieveBotMessage(this, "That IP address is not banned.");
	}
    }

    /**
     * Helper function that both ban commands use. Both ban and banrange commands
     * call this, but with a different range argument.
     * @param client
     * @param args
     * @param range To perform a ban on the ip range of the target.
     */
    private void banningAux(BotClient client, String[] args, boolean range) {
        if (! identifiedRequiredHelper(client)) {
            return;
        }

        // Reply with usage if malformed
        if (args.length != 2) {
            try {
                client.recieveBotMessage(this, help.getCommandUsage("ban"));
            } catch (CommandNotFoundException e) {
                client.recieveBotMessage(this, "Invalid number of args.");
            }
            return;
        }

        // Parse the arguments
        String channelName = args[0];
        int num;

        try {
            num = Integer.parseInt(args[1]);
        } catch (NumberFormatException e) {
            client.recieveBotMessage(this, "Invalid number.");
            return;
        }

        // Retrieve the channel, failing on bad channel name
        Channel channel;

        try {
            channel = server.getChannel(channelName);
        } catch (IllegalChannelNameException e) {
            client.recieveBotMessage(this, "Invalid channel.");
            return;
        }

        // Allow the user to kick only if they have permission
        AuthedAccount account = identifiedClients.get(client);
        if (! PermissionManager.canBan(account, channelName)) {
            client.recieveBotMessage(this, "You do not have sufficient privileges to ban on this channel.");
            return;
        }
	
        // Perform the ban
        try {
            channel.ban(num, range);
        } catch (ChannelParticipantNotFoundException e) {
            String errorMsg = "The provided number was either too old or has not yet occurred.";
            client.recieveBotMessage(this, errorMsg);
            return;
        }

        client.recieveBotMessage(this, "Ban issued.");
    }

    /**
     * The format of this command is:
     *   BAN channel num
     */
    public void handleBan(BotClient client, String[] args) {
	banningAux(client, args, false);
    }

    /**
     * The format of this command is:
     *   BANRANGE channel num
     */
    public void handleBanRange(BotClient client, String[] args) {
	banningAux(client, args, true);
    }


    /**
     * The format of this command is:
     *   CLAIM channel
     */
    public void handleClaim(BotClient client, String[] args) {
        if (! identifiedRequiredHelper(client)) {
            return;
        }

        // Reply with usage and return if malformed
        if (args.length != 1) {
            try {
                client.recieveBotMessage(this, help.getCommandUsage("claim"));
            } catch (CommandNotFoundException e) {
                client.recieveBotMessage(this, "Invalid number of args.");
            }
            return;
        }

        // Check to make sure channel name is valid and that channel is empty
        String channelName = args[0];
        Channel channel;

        try {
            channel = server.getChannel(channelName);
        } catch (IllegalChannelNameException e) {
            String message = "Invalid channel name.";
            client.recieveBotMessage(this, message);
            return;
        }

        if (channel.getNumParticipants() > 0) {
            String message = "You can only claim empty channels.";
            client.recieveBotMessage(this, message);
            return;
        }

        // Check if someone already owns the channel.
        if (server.config.checkChannelOwned(channelName)) {
            String message = "That channel is already claimed.";
            client.recieveBotMessage(this, message);
            return;
        }

        // Go ahead and give this user ownership of the channel.
        AuthedAccount account = identifiedClients.get(client);
        account.setChannelPriviledge(channelName, ChannelPrivilege.OWNER);

        server.config.saveAuthedAccount(account);

        String message = "You now have privilege "
            + ChannelPrivilege.OWNER.name().toUpperCase()
            + " on " + channelName;
        client.recieveBotMessage(this, message);
    }

    /**
     * The format of this command is:
     *    KICK channel num
     */
    public void handleKick(BotClient client, String[] args) {
        if (! identifiedRequiredHelper(client)) {
            return;
        }

        // Reply with usage and return if malformed
        if (args.length != 2) {
            String kickUsage = "Usage: KICK <channel> <num>";
            client.recieveBotMessage(this, kickUsage);
            return;
        }

        // Parse the arguments
        String channelName = args[0];
        int num;

        try {
            num = Integer.parseInt(args[1]);
        } catch (NumberFormatException e) {
            String errorMsg = "Invalid number.";
            client.recieveBotMessage(this, errorMsg);
            return;
        }

        // Retrieve the channel, failing on bad channel name
        Channel channel;

        try {
            channel = server.getChannel(channelName);
        } catch (IllegalChannelNameException e) {
            client.recieveBotMessage(this, "Invalid channel.");
            return;
        }

        // Allow the user to kick only if they have permission
        AuthedAccount account = identifiedClients.get(client);
        if (! PermissionManager.canKick(account, channelName)) {
            String errorMsg = "You do not have sufficient privileges to kick on this channel.";
            client.recieveBotMessage(this, errorMsg);
            return;
        }

        // Perform the kick
        try {
            channel.kick(num);
        } catch (ChannelParticipantNotFoundException e) {
            String errorMsg = "The provided number was either too old or has not yet occurred.";
            client.recieveBotMessage(this, errorMsg);
            return;
        }

        client.recieveBotMessage(this, "Kick issued.");
    }

    /**
     * Handles the notice command. 
     * @param args Must not be zero-length array.
     */
    public void handleNotice(BotClient client, String[] args) {
        if (! identifiedRequiredHelper(client))
            return;
	
	// Do the work
        AuthedAccount account = identifiedClients.get(client);
	String message = Utils.join(" ", args);

	if (message.isEmpty()) {
	    client.recieveBotMessage(this, "Must specify a message to send.");
	} else if (PermissionManager.canNotice(account)) {
	    server.sendSystemNotice(message);
	    client.recieveBotMessage(this,  "Notice sent.");
	} else {
            client.recieveBotMessage(this,  "Insufficient privileges.");
        }
    }

    /**
     * The format of this command is
     *    REGISTER username password
     */
    public void handleRegister(BotClient client, String[] args) {
        // Deny access to registering account if the user is already identified.
        if (identifiedClients.containsKey(client)) {
            String message = "You cannot use the REGISTER command while logged in.";
            client.recieveBotMessage(this, message);
            return;
        }

        // Reply with usage if malformed
        if (args.length != 2) {
            String registerUsage = "Usage: REGISTER <username> <password>";
            client.recieveBotMessage(this, registerUsage);
            return;
        }

        // Reply with error if an account already exists with this username
        String username = args[0];
        String password = args[1];

        if (server.config.getAuthedAccount(username) != null) {
            String message = "An account already exists with that username.";
            client.recieveBotMessage(this, message);
            return;
        }

        // Go ahead and make and store the account.
        AuthedAccount account;

        try {
            account = new AuthedAccount(username, ConfigUtils.hashPassword(password), false);
        } catch (NoSuchAlgorithmException e) {
            String message = "An error occured.";
            client.recieveBotMessage(this, message);
            return;
        }

        server.config.saveAuthedAccount(account);
        identifiedClients.put(client, account);
        client.recieveBotMessage(this, "Account successfully registered. You are now logged in.");
    }

    /**
     * The format of this command is:
     *    REVEAL channel
     */
    public void handleReveal(BotClient client, String[] args) {
        if (! identifiedRequiredHelper(client)) {
            return;
        }

        // Reply with usage if malformed
        if (args.length != 1) {
            try {
                client.recieveBotMessage(this, help.getCommandUsage("reveal"));
            } catch (CommandNotFoundException e) {
                client.recieveBotMessage(this, "Invalid number of args.");
            }
            return;
        }

        // Attempt to retrieve the channel, halt on failure.
        String channelName = args[0];
        Channel channel;

        try {
            channel = server.getChannel(channelName);
        } catch (IllegalChannelNameException e) {
            client.recieveBotMessage(this, "Invalid channel.");
            return;
        }

        // Halt if the client does not have correct permission to reveal on this channel.
        AuthedAccount account = identifiedClients.get(client);
        if (! PermissionManager.canReveal(account, channelName)) {
            client.recieveBotMessage(this, "You do not have sufficient privileges to reveal on this channel.");
            return;
        }

        // Send reveal message
        client.recieveBotMessage(this, "Start of recent messages");
        for (ChannelMessage recentMessage : channel.getRecentMessages()) {
            client.recieveBotMessage(this, "   [" + recentMessage.num + "] " + recentMessage.message);
        }
        client.recieveBotMessage(this, "End of recent messages");
    }

    public void handleUnknown(BotClient client, String command) {
        client.recieveBotMessage(this, "Unknown command " + command + ". Type HELP for a list of commands.");
    }

    public String getNick() {
        return "AdminServ";
    }
}
