package net.meltingwax.anonircd.conf;

import net.meltingwax.anonircd.*;
import net.meltingwax.anonircd.conf.excepts.*;
import net.meltingwax.anonircd.i.*;
import java.io.*;
import java.security.*;
import java.sql.*;
import java.util.*;


/**
 * Provides Configuration settings through a configuration directory. Usually,
 * this directory is ~/.anonircd.
 *
 * The contents of a configuration directory are:
 * - a sqlite3 database (@see DB_FILE_NAME)
 * - a properties file for general settings (@see PROP_FILE_NAME)
 * - a text file for the MOTD (@see MOTD_FILE_NAME)
 */
public class ConfigDir implements ConfigProvider {
    private Properties prop;
    private ConfigProvider defaultConfig;
    private String configDirPath;
    private Connection dbConn;
    private Map<String, AuthedAccount> authedAccountPool;  // so we only have one AuthedAccount instance per username

    public final static String PROP_FILE_NAME = "anonircd.properties";
    public final static String MOTD_FILE_NAME = "motd.txt";
    public final static String DB_FILE_NAME = "anonircd.db";

    /**
     * Constructor
     * @param configDirPath (The path to the configuration directory)
     * @param defaultConfig (A ConfigProvider to use for default values)
     * @throws ConfigLoadException (An error occurred loading the config
     * directory).
     */
    public ConfigDir(String configDirPath, ConfigProvider defaultConfig)
        throws ConfigLoadException {

        this.configDirPath = configDirPath;
        this.defaultConfig = defaultConfig;
        this.authedAccountPool = new HashMap<String, AuthedAccount>();

        reload();

        assert ConfigUtils.canHashPasswords();
    }

    /**
     * Generates a starting config file.
     * @param configDirPath The configuration directory
     */
    public static void genStarter(String configDirPath) throws ConfigGenException {
        Console console = System.console();

        if (console == null) {
            throw new ConfigGenException("Unable to get console for input", new NullPointerException());
        }

        System.out.print("Sysop Username: ");
        String username = console.readLine();

        System.out.print("Sysop Password: ");
        String password = new String(console.readPassword());

        genStarter(configDirPath, username, password);
    }

    /**
     * Generates a starting config file.
     * @param configDirPath The configuration directory
     * @param username Username of the first SysOp
     * @param passHash Password of the first SysOp
     */
    public static void genStarter(String configDirPath, String username, String password) throws ConfigGenException {
        // Create dir if doesn't exist
        File dir = new File(configDirPath);

        if (dir.exists()) {
            throw new ConfigGenException("Configuration directory at " + dir.getPath() + " already exists.",
                                         new IllegalArgumentException());
        }

        dir.mkdirs();

        // Generate the .properties file
        File propFile = new File(configDirPath, PROP_FILE_NAME);
        Properties starter = new Properties();
        starter.setProperty("port", "6667");

        try {
            starter.store(new FileOutputStream(propFile), "");
        } catch (IOException e) {
            throw new ConfigGenException("Error writing to " + PROP_FILE_NAME, e);
        }

        // Generate the MOTD file
        try {
            File motdFile = new File(configDirPath, MOTD_FILE_NAME);
            FileOutputStream motdOut = new FileOutputStream(motdFile);
            motdOut.write("Welcome to AnonIRCd\n".getBytes());
            motdOut.close();
        } catch (IOException e) {
            throw new ConfigGenException("Error writing to " + MOTD_FILE_NAME, e);
        }

        // Generate the database
        try {
            Class.forName("org.sqlite.JDBC");
        } catch (ClassNotFoundException e) {
            throw new ConfigGenException("Error initializing SqliteJDBC"
                                         + " (http://www.xerial.org/trac/Xerial/wiki/SQLiteJDBC). Is it in"
                                         + " your classpath? ", e);
        }

        File dbFile = new File(configDirPath, DB_FILE_NAME);

        Connection conn;
        try {
            conn = DriverManager.getConnection("jdbc:sqlite:" + dbFile.getPath());

            Statement createUsersTable = conn.createStatement();
            createUsersTable.execute("CREATE TABLE users ("
                                     + "username TEXT UNIQUE, "
                                     + "pass_hash TEXT, "
                                     + "is_sysop INT "
                                     + ");");

            Statement createChannelsTable = conn.createStatement();
            createChannelsTable.execute("CREATE TABLE channels ("
                                        + "name TEXT, "
                                        + "username TEXT, "
                                        + "level INT "
                                        + ");");

            Statement createChannelsMetaDataTable = conn.createStatement();
            createChannelsMetaDataTable.execute("CREATE TABLE channel_meta ("
                                                + "name TEXT, "
                                                + "topic TEXT "
                                                + ");");

            Statement createBansTable = conn.createStatement();
            createBansTable.execute("CREATE TABLE bans ("
                                    + "channel_name TEXT, "
                                    + "ip_addr TEXT, "
                                    + "time INTEGER "
                                    + ");");
        } catch (SQLException e) {
            e.printStackTrace();
            throw new ConfigGenException("Error initializing " + DB_FILE_NAME, e);
        }

        // Add a superuser to the database
        String passHash;
        try {
            passHash = ConfigUtils.hashPassword(password);
        } catch (NoSuchAlgorithmException e) {
            throw new ConfigGenException("SHA-1 not available", e);
        }

        try {
            PreparedStatement prep = conn.prepareStatement("INSERT INTO users VALUES (?, ?, ?);");
            prep.setString(1, username);
            prep.setString(2, passHash);
            prep.setInt(3, 1);
            prep.execute();
        } catch (SQLException e) {
            throw new ConfigGenException("Error saving op account", e);
        }
    }

    /**
     * Reloads the configuration directory.
     */
    public void reload() throws ConfigLoadException {

        // Reload the properties file
        prop = new Properties();

        File propFile = new File(configDirPath, PROP_FILE_NAME);
        FileInputStream propStream;

        try {
            propStream = new FileInputStream(propFile);
        } catch (IOException e) {
            throw new ConfigLoadException(e);
        }

        try {
            prop.load(propStream);
        } catch (IOException e) {
            throw new ConfigLoadException(e);
        }

        // Reconnect to the database
        try {
            Class.forName("org.sqlite.JDBC");
            String path = new File(configDirPath, DB_FILE_NAME).getPath();
            dbConn = DriverManager.getConnection("jdbc:sqlite:" + path);
        } catch (Exception e) {
            throw new ConfigLoadException(e);
        }
    }

    /**
     * Verify that this login/password combination indeed matches that of
     * a stored authorized account.
     * @param username The username of the account
     * @param password  The password of the account
     */
    public boolean checkAuthCredentials(String username, String password) {
        String encryptedPass;

        try {
            encryptedPass = ConfigUtils.hashPassword(password);
        } catch (NoSuchAlgorithmException e) {
            // We checked for ability to encrypt upon construction, so we
            // shouldn't get here. However, let's fail just in case.
            return false;
        }

        try {
            PreparedStatement stat = dbConn.prepareStatement("SELECT * FROM users WHERE username = ? AND pass_hash = ?;");
            stat.setString(1, username);
            stat.setString(2, encryptedPass);

            ResultSet rs = stat.executeQuery();

            while (rs.next()) {
                return true;
            }

            return false;

        } catch (SQLException e) {
            e.printStackTrace();
            return false;
        }

    }


    /**
     * Returns an AuthedAccount instance for the given username, or null if one
     * does not exist. Care is taken to ensure that there is only one instance
     * of an AuthedAccount per username.
     */
    public AuthedAccount getAuthedAccount(String username) {
        return getAuthedAccount(username, false);
    }

    /**
     * Returns an AuthedAccount from the database, optionally skipping the pool.
     * Do NOT use this method except for unit testing. Using this method in
     * application code WILL cause an inconsistent state.
     */
    public AuthedAccount getAuthedAccount(String username, boolean skipPool) {
        if (! skipPool && authedAccountPool.containsKey(username)) {
            return authedAccountPool.get(username);
        }

        try {
            PreparedStatement usersStat = dbConn.prepareStatement("SELECT * FROM users WHERE username = ?;");
            usersStat.setString(1, username);

            ResultSet usersRS = usersStat.executeQuery();

            boolean rowExists = usersRS.next();

            if (rowExists) {
                AuthedAccount account = new AuthedAccount(username, usersRS.getString("pass_hash"), usersRS.getInt("is_sysop") == 1);

                PreparedStatement channelsStat = dbConn.prepareStatement("SELECT * FROM channels WHERE username = ?;");
                channelsStat.setString(1, username);

                ResultSet channelsRS = channelsStat.executeQuery();
                while (channelsRS.next()) {
                    account.setChannelPriviledge(channelsRS.getString("name"), levelIntToChannelPrivilege(channelsRS.getInt("level")));
                }

                authedAccountPool.put(username, account);

                return account;
            } else {
                return null;
            }
        } catch (SQLException e) {
            return null;
        }
    }

    /**
     * Returns the MOTD String (with possible \n's)
     */
    public String getMOTD() {
        File motdFile = new File(configDirPath, MOTD_FILE_NAME);

        // Create a scanner over the file
        Scanner scanner;

        try {
            scanner = new Scanner(motdFile);
        } catch (FileNotFoundException e) {
            return defaultConfig.getMOTD();
        }

        // Convert to StringBuffer
        StringBuffer motd = new StringBuffer();

        while (scanner.hasNextLine()) {
            motd.append(scanner.nextLine() + "\n");
        }

        motd.deleteCharAt(motd.length() - 1);

        // Return as string
        return motd.toString();
    }

    /**
     * Returns the maximum number of recent channel messages to save for the REVEAL
     * command.
     */
    public int getMaxChannelRecentMessages() {
        String key = "max_channel_recent_messages";
        int n;

        if (prop.keySet().contains(key)) {
            Integer nInt;
            try {
                nInt = new Integer(prop.getProperty(key));
                n = nInt.intValue();
            } catch (NumberFormatException e) {
                n = defaultConfig.getMaxChannelRecentMessages();
            }
        } else {
            n = defaultConfig.getMaxChannelRecentMessages();
        }

        return n;
    }

    /**
     * Returns the maximum length of a message.
     */
    public int getMaxMessageLength() {
        String key = "max_message_length";
        int n;

        if (prop.keySet().contains(key)) {
            Integer nInt;
            try {
                nInt = new Integer(prop.getProperty(key));
                n = nInt.intValue();
            } catch (NumberFormatException e) {
                n = defaultConfig.getMaxMessageLength();
            }
        } else {
            n = defaultConfig.getMaxMessageLength();
        }

        return n;
    }

    /**
     * Returns the maximum length of a topic.
     */
    public int getMaxTopicLength() {
        String key = "max_topic_length";
        int n;

        if (prop.keySet().contains(key)) {
            Integer nInt;
            try {
                nInt = new Integer(prop.getProperty(key));
                n = nInt.intValue();
            } catch (NumberFormatException e) {
                n = defaultConfig.getMaxTopicLength();
            }
        } else {
            n = defaultConfig.getMaxTopicLength();
        }

        return n;
    }

    /**
     * Returns the port to listen on.
     */
    public int getPort() {
        String key = "port";
        int n;

        if (prop.keySet().contains(key)) {

            Integer nInt;
            try {
                nInt = new Integer(prop.getProperty(key));
                n = nInt.intValue();
            } catch (NumberFormatException e) {
                e.printStackTrace();
                n = defaultConfig.getPort();
            }
        } else {
            n = defaultConfig.getPort();
        }

        return n;
    }

    /**
     * Converts a level integer (as present in the users database) to a
     * corresponding ChannelPrivilege value.
     * @param level (The level int to convert)
     * @return (The corresponding ChannelPriviledge)
     */
    public static ChannelPrivilege levelIntToChannelPrivilege(int level) {
        ChannelPrivilege privilege;

        if (level == 0) {
            privilege = ChannelPrivilege.NONE;
        } else if (level == 100) {
            privilege = ChannelPrivilege.HALF_OP;
        } else if (level == 200) {
            privilege = ChannelPrivilege.OP;
        } else if (level == 300) {
            privilege = ChannelPrivilege.OWNER;
        } else {
            throw new IllegalArgumentException();
        }

        return privilege;
    }

    /**
     * Converts a ChannelPrivilege value to a level (integer as stored in the
     * database. This is the inverse function of
     * @param privilege (The ChannelPrivilege to convert)
     * @return (The corresponding level int)
     */
    public static int channelPrivelegeToLevelInt(ChannelPrivilege privilege) {
        if (privilege == ChannelPrivilege.NONE) {
            return 0;
        } else if (privilege == ChannelPrivilege.HALF_OP) {
            return 100;
        } else if (privilege == ChannelPrivilege.OP) {
            return 200;
        } else if (privilege == ChannelPrivilege.OWNER) {
            return 300;
        } else {
            throw new UnsupportedOperationException();
        }
    }

    /**
     * Saves an AuthedAccount. If an account already exists with this
     * username, it is overwritten.
     */
    public void saveAuthedAccount(AuthedAccount account) {
        try {
            synchronized (dbConn) {
                // Update the users table. We branch out and do a UPDATE vs INSERT based on
                // whether we are overwriting.
                PreparedStatement usersPresence = dbConn.prepareStatement("SELECT * FROM users WHERE username = ?;");
                usersPresence.setString(1, account.getUsername());

                boolean usersRowExists = usersPresence.executeQuery().next();

                if (usersRowExists) {
                    String query = "UPDATE users SET pass_hash = ?, is_sysop = ? WHERE username = ?;";
                    PreparedStatement updateUsers = dbConn.prepareStatement(query);
                    updateUsers.setString(1, account.getPasswordHash());
                    updateUsers.setInt(2, account.getIsSysOp() ? 1 : 0);
                    updateUsers.setString(3, account.getUsername());
                    updateUsers.execute();
                } else {
                    PreparedStatement addToUsers = dbConn.prepareStatement("INSERT INTO users VALUES (?, ?, ?);");
                    addToUsers.setString(1, account.getUsername());
                    addToUsers.setString(2, account.getPasswordHash());
                    addToUsers.setInt(3, account.getIsSysOp() ? 1 : 0);
                    addToUsers.execute();
                }

                // Update the channels table. We first query for what is there, and then
                // branch to UPDATE vs INSERT on a row-by-row basis based on that.
                PreparedStatement channelsQuery = dbConn.prepareStatement("SELECT * FROM channels WHERE username = ?;");
                channelsQuery.setString(1, account.getUsername());
                ResultSet channelQueryResults = channelsQuery.executeQuery();

                Set<String> existingChannelRows = new HashSet<String>();
                while (channelQueryResults.next()) {
                    existingChannelRows.add(channelQueryResults.getString("name"));
                }

                for (String channelName : account.getChannelsWithElevatedPrivileges()) {
                    if (existingChannelRows.contains(channelName)) {
                        String query = "UPDATE channels SET level = ? WHERE name = ? AND username = ?";
                        PreparedStatement updateChannels = dbConn.prepareStatement(query);
                        updateChannels.setInt(1, channelPrivelegeToLevelInt(account.getChannelPrivilege(channelName)));
                        updateChannels.setString(2, channelName);
                        updateChannels.setString(3, account.getUsername());
                        updateChannels.execute();
                    } else {
                        PreparedStatement addToChannels = dbConn.prepareStatement("INSERT INTO channels VALUES (?, ?, ?);");
                        addToChannels.setString(1, channelName);
                        addToChannels.setString(2, account.getUsername());
                        addToChannels.setInt(3, channelPrivelegeToLevelInt(account.getChannelPrivilege(channelName)));
                        addToChannels.execute();
                    }
                }
            }
        } catch (SQLException e) {
            e.printStackTrace();
        }

        synchronized (authedAccountPool) {
            if (! authedAccountPool.containsKey(account.getUsername())) {
                authedAccountPool.put(account.getUsername(), account);
            }
        }
    }

    /**
     * Check whether an IP is banned from a given channel.
     * @param channelName name of the channel, or null to select server-wide bans only.
     * @param ipAddr IP address of the user.
     */
    public boolean checkForChannelBan(String channelName, String ipAddr) {
        if (channelName == null)
            channelName = "*";

        channelName = channelName.toLowerCase();

        try {
            String q =
                "SELECT * FROM bans"
                + " WHERE (channel_name = ? OR channel_name = '*')"
                + " AND ? LIKE ip_addr;";
            PreparedStatement stat = dbConn.prepareStatement(q);
            stat.setString(1, channelName);
            stat.setString(2, ipAddr);
            return stat.executeQuery().next();
        } catch (SQLException e) {
            e.printStackTrace();
            throw new RuntimeException(e);
        }
    }

    public boolean checkForTrustedWebClient(String clientIP) {
        String key = "trusted_webclients";

        if (prop.keySet().contains(key)) {
            String value = prop.getProperty(key);
            String[] trustedIPs = value.trim().split(",");

            for (String trustedIP : trustedIPs) {
                if (trustedIP.equals(clientIP)) {
                    return true;
                }
            }

            return false;
        } else {
            return defaultConfig.checkForTrustedWebClient(clientIP);
        }
    }

    /**
     * Adds a stored IP ban on a channel.
     *
     * Adds a row to the `bans` table if one does not already exist.
     */
    public void addChannelBan(String channelName, String ipAddr) {
        channelName = channelName.toLowerCase();


        if (! checkForChannelBan(channelName, ipAddr)) {
            try {
                PreparedStatement stat = dbConn.prepareStatement("INSERT INTO bans VALUES (?, ?, ?);");
                stat.setString(1, channelName);
                stat.setString(2, ipAddr.replace("*", "%"));
                stat.setInt(3, (int)(System.currentTimeMillis()/1000));
                stat.execute();
            } catch (SQLException e) {
                e.printStackTrace();
            }
        }
    }


    /**
     * Removes a stored IP ban on a channel.
     *
     * Removes a DELETE on the bans table with suitable WHERE clause.
     */
    public void removeChannelBan(String channelName, String ipAddr) {
        channelName = channelName.toLowerCase();

        String query = "DELETE FROM bans WHERE channel_name = ? AND ip_addr =?";

        try {
            PreparedStatement stat = dbConn.prepareStatement(query);
            stat.setString(1, channelName);
            stat.setString(2, ipAddr);
            stat.execute();
        } catch (SQLException e) {
            e.printStackTrace();
        }
    }

    public boolean checkChannelOwned(String channelName) {
        channelName = channelName.toLowerCase();

        try {
            String query = "SELECT * FROM channels WHERE name = ? AND level = ?;";

            PreparedStatement stat = dbConn.prepareStatement(query);
            stat.setString(1, channelName);
            stat.setInt(2, channelPrivelegeToLevelInt(ChannelPrivilege.OWNER));

            ResultSet results = stat.executeQuery();
            while (results.next()) {
                return true;
            }
            return false;
        } catch (SQLException e) {
            e.printStackTrace();
            return true;
        }
    }

    public String getChannelTopic(String channelName) {
        try {
            PreparedStatement statement = dbConn.prepareStatement("SELECT * FROM channel_meta WHERE name = ?;");
            statement.setString(1, channelName);
            ResultSet results = statement.executeQuery();

            while(results.next()) {
                return results.getString("topic");
            }

            return null;
        } catch (SQLException e) {
            e.printStackTrace();
            return null;
        }
    }

    public void saveChannelTopic(String channelName, String topic) {
        try {
            PreparedStatement statement = dbConn.prepareStatement("SELECT * FROM channel_meta WHERE name = ?;");
            statement.setString(1, channelName);
            ResultSet results = statement.executeQuery();
            boolean rowExists = results.next();

            if (rowExists) {
                PreparedStatement stat = dbConn.prepareStatement("UPDATE channel_meta SET topic = ? WHERE name = ?;");
                stat.setString(1, topic);
                stat.setString(2, channelName);
                stat.execute();
            } else {
                PreparedStatement stat = dbConn.prepareStatement("INSERT INTO channel_meta VALUES (?, ?);");
                stat.setString(1, channelName);
                stat.setString(2, topic);
                stat.execute();
            }
        } catch (SQLException e) {
            e.printStackTrace();
        }
    }
}
