package net.meltingwax.anonircd.conf.tests;

import net.meltingwax.anonircd.*;
import net.meltingwax.anonircd.conf.*;
import net.meltingwax.anonircd.conf.excepts.*;
import net.meltingwax.anonircd.i.*;
import junit.framework.*;
import java.io.*;
import java.security.*;


/**
 * Test cases for the ConfigDir class.
 *
 * This does a few hacks to make a config dir on the fly for each test. See the make() and
 * make(String[]) methods. The latter makes the config directory with a custom properties file.
 */
public class ConfigDirTest extends TestCase {
    private final String configDirPath = "/tmp/.anonird_test";
    private final String sysOpUsername = "admin";
    private final String sysOpPassword = "098dfasaro8anjas9whyi7aw";

    private ConfigDir make() {
        // Removed testing dir if exists.
        File dir = new File(configDirPath);
        if (dir.exists()) {
            deleteFile(dir);
        }

        // Generate a starting directory
        try {
            ConfigDir.genStarter(configDirPath, sysOpUsername, sysOpPassword);
        } catch (ConfigGenException e) {
            e.printStackTrace();
            fail();
        }

        // Create ConfigDir instance
        ConfigDir config;

        try {
            config = new ConfigDir(configDirPath, new ConfigDefaults());
        } catch (ConfigLoadException e) {
            e.printStackTrace();
            fail();
            return null;
        }

        return config;
    }

    private ConfigDir make(String[] lines) {
        // Removed testing dir if exists.
        File dir = new File(configDirPath);
        if (dir.exists()) {
            deleteFile(dir);
        }

        // Generate a starting directory
        try {
            ConfigDir.genStarter(configDirPath, sysOpUsername, sysOpPassword);
        } catch (ConfigGenException e) {
            e.printStackTrace();
            fail();
            return null;
        }

        // Append extra lines to properties file
        try {
            FileWriter fstream = new FileWriter(new File(configDirPath, ConfigDir.PROP_FILE_NAME).getPath(), true);
            BufferedWriter out = new BufferedWriter(fstream);

            for (String line: lines) {
                out.write(line.toCharArray());
                out.write('\n');
            }
            out.close();
        } catch (IOException e) {
            e.printStackTrace();
            fail();
            return null;
        }

        // Create ConfigDir instance
        ConfigDir config;

        try {
            config = new ConfigDir(configDirPath, new ConfigDefaults());
        } catch (ConfigLoadException e) {
            e.printStackTrace();
            fail();
            return null;
        }
        return config;
    }

    private void deleteFile(File f) {
        if (f.isDirectory()) {
            for (File c : f.listFiles()) {
                deleteFile(c);
            }
        }
        f.delete();
    }

    /**
     * Test basic ability to store bans.
     */
    public void testChannelBan_basic() {
        ConfigProvider config;

        config = make();
        assertFalse(config.checkForChannelBan("#lounge", "127.0.0.1"));
        config.addChannelBan("#lounge", "127.0.0.1");
        assertTrue(config.checkForChannelBan("#lounge", "127.0.0.1"));
	
        config = make();
        config.addChannelBan("#lounge", "127.0.0.1");
        assertFalse(config.checkForChannelBan("#lounge", "127.0.0.2"));
    }

    /**
     * Test storing and unstoring bans.
     */
    public void testChannelBan_addRemove() {
	ConfigProvider config;
	String ipA = "127.0.0.1";
	String ipB = "128.0.0.1";
	String chan = "#jp";

	config = make();
	assertFalse(config.checkForChannelBan(chan, ipA));
	assertFalse(config.checkForChannelBan(chan, ipB));

	config.addChannelBan(chan, ipA);
	assertTrue(config.checkForChannelBan(chan, ipA));
	assertFalse(config.checkForChannelBan(chan, ipB));
	
	config.removeChannelBan(chan, ipA);
	assertFalse(config.checkForChannelBan(chan, ipA));
	assertFalse(config.checkForChannelBan(chan, ipB));

	config.addChannelBan(chan, ipA);
	config.addChannelBan(chan, ipB);       
	assertTrue(config.checkForChannelBan(chan, ipA));
	assertTrue(config.checkForChannelBan(chan, ipB));
    }

    public void testChannelBanRange() {
        ConfigProvider config;

        config = make();
        assertFalse(config.checkForChannelBan("#lounge", "127.0.0.1"));
        config.addChannelBan("#lounge", "127.0.*.*");
        assertTrue(config.checkForChannelBan("#lounge", "127.0.0.1"));
    }

    public void testChannelPrivelegeToLevelInt() {
        for (ChannelPrivilege p : ChannelPrivilege.values()) {
            assertTrue(ConfigDir.channelPrivelegeToLevelInt(p) >= 0);
        }
    }

    public void testPrivilegeLevelIntInverse() {
        for (ChannelPrivilege p : ChannelPrivilege.values()) {
            assertEquals(p, ConfigDir.levelIntToChannelPrivilege(ConfigDir.channelPrivelegeToLevelInt((p))));
        }
    }

    public void testCheckAuthCredentials() {
        ConfigProvider config = make();

	// Test on the SysOp account we know exists
        assertTrue(config.checkAuthCredentials(sysOpUsername, sysOpPassword));
        assertFalse(config.checkAuthCredentials(sysOpUsername, ""));
        assertFalse(config.checkAuthCredentials(sysOpUsername, "abc"));

	// Test on an account we know does not exist
        assertFalse(config.checkAuthCredentials("abc", "def"));
        assertFalse(config.checkAuthCredentials("", ""));
    }
    
    public void testGetAuthedAccount() {
        ConfigProvider config = make();
        AuthedAccount account;

        account = config.getAuthedAccount(sysOpUsername);
        assertNotNull(account);
        assertNotNull(account.getUsername());
        assertTrue(account.getUsername().equals(sysOpUsername));
        assertTrue(account.getIsSysOp());
        assertTrue(account.getChannelPrivilege("#abc") == ChannelPrivilege.NONE);

        account = config.getAuthedAccount(sysOpUsername + "abc");
        assertNull(account);

        // Check multiple calls return the same instance
        assertTrue(config.getAuthedAccount(sysOpUsername) == config.getAuthedAccount(sysOpUsername));
    }

    public void testGetMaxChannelRecentMessages_basic() {
        ConfigProvider config = make();

        assertTrue(config.getMaxChannelRecentMessages() > 0);	
    }

    public void testGetMaxChannelRecentMessages_readsPropFile() {
	int num = 123456;
        String[] lines = {"max_channel_recent_messages=" + num};

        ConfigProvider config = make(lines);

        assertEquals(num, config.getMaxChannelRecentMessages());
    }

    public void testGetMaxMessageLength() {
        ConfigProvider config;

        config = make();
        assertTrue(config.getMaxMessageLength() > 0);

        String[] lines = {"max_message_length=123456"};
        config = make(lines);
        assertEquals(123456, config.getMaxMessageLength());
    }

    public void testGetMaxTopicLength() {
        ConfigProvider config;

        config = make();
        assertTrue(config.getMaxTopicLength() > 0);

        String[] lines = {"max_topic_length=123456"};
        config = make(lines);
        assertEquals(123456, config.getMaxTopicLength());
    }

    public void testGetMOTD() {
        ConfigProvider config = make();

        assertNotNull(config.getMOTD());

        // Now, we will overwrite the MOTD file without reloading. We want
        // to test that getMOTD() reads the file on every call, and does
        // not need a call to reload() in order for changes to take effect.
        String newMOTD = config.getMOTD() + "Hello world\n\this is my MTOD";

        try {
            FileWriter fstream = new FileWriter(new File(configDirPath, ConfigDir.MOTD_FILE_NAME).getPath());
            BufferedWriter out = new BufferedWriter(fstream);
            out.write(newMOTD);
            out.close();
        } catch (IOException e) {
            e.printStackTrace();
            fail();
        }

        assertTrue(config.getMOTD().equals(newMOTD));
    }

    public void testGetPort() {
        ConfigProvider config;

        config = make();
        assertTrue(config.getPort() > 0);

        String[] lines = {"port=123456"};
        config = make(lines);
        assertEquals(123456, config.getPort());
    }

    public void testSaveAuthedAccount_removeSysOp() {
        ConfigDir config;
        AuthedAccount account;

        // Test removing SysOp status
        config = make();
        account = config.getAuthedAccount(sysOpUsername);
        assertTrue(account.getIsSysOp());
        account.setIsSysOp(false);
        config.saveAuthedAccount(account);

        account = config.getAuthedAccount(sysOpUsername, true);
        assertFalse(account.getIsSysOp());
        assertEquals(0, account.getChannelsWithElevatedPrivileges().size());
    }

    public void testSaveAuthedAccount_addOneChannel() {
        ConfigDir config;
        AuthedAccount account;

        // Test adding one channel for the account
        config = make();
        account = config.getAuthedAccount(sysOpUsername);
        assertEquals(0, account.getChannelsWithElevatedPrivileges().size());
        account.setChannelPriviledge("#lounge", ChannelPrivilege.HALF_OP);
        assertEquals(1, account.getChannelsWithElevatedPrivileges().size());
        config.saveAuthedAccount(account);

        account = config.getAuthedAccount(sysOpUsername, true);
        assertEquals(1, account.getChannelsWithElevatedPrivileges().size());
        assertEquals(ChannelPrivilege.HALF_OP, account.getChannelPrivilege("#lounge"));
    }

    public void testSaveAuthedAccount_removeSomeChannels() throws ConfigLoadException {
        ConfigDir config;
        AuthedAccount account;

        // * Starts with three channels, saves.
        // * Removes two, saves.
        // * Checks only one is left
        config = make();

        account = new AuthedAccount("joe", "aaaaa", false);
        account.setChannelPriviledge("#lounge", ChannelPrivilege.OWNER);
        account.setChannelPriviledge("#4ch", ChannelPrivilege.OP);
        account.setChannelPriviledge("#dqn", ChannelPrivilege.HALF_OP);
        assertEquals(3, account.getChannelsWithElevatedPrivileges().size());
        config.saveAuthedAccount(account);
        config.reload();  // makes new connection instance

        account = config.getAuthedAccount("joe", true);
        assertEquals(3, account.getChannelsWithElevatedPrivileges().size());
        assertEquals(ChannelPrivilege.OWNER, account.getChannelPrivilege("#lounge"));
        assertEquals(ChannelPrivilege.OP, account.getChannelPrivilege("#4ch"));
        assertEquals(ChannelPrivilege.HALF_OP, account.getChannelPrivilege("#dqn"));

        account.setChannelPriviledge("#4ch", ChannelPrivilege.NONE);
        account.setChannelPriviledge("#dqn", ChannelPrivilege.NONE);
        config.saveAuthedAccount(account);
        config.reload();  // makes new connection instance

        assertEquals(1, account.getChannelsWithElevatedPrivileges().size());
        assertEquals(ChannelPrivilege.OWNER, account.getChannelPrivilege("#lounge"));
        assertEquals(ChannelPrivilege.NONE, account.getChannelPrivilege("#4ch"));
        assertEquals(ChannelPrivilege.NONE, account.getChannelPrivilege("#dqn"));
    }

    public void testGetChannelTopic_nothingSaved() {
        ConfigDir config = make();

        assertNull(config.getChannelTopic("#lounge"));
        assertNull(config.getChannelTopic("#jp"));
    }

    public void testSaveChannelTopic() {
        ConfigDir config = make();

        String[] channels = {"#jp", "#lounge", "#dev"};
        String[] topics = {"", "Hello world", "HaHlEOal 8 2l a8s das;d // a2!"};

        for (String channel : channels) {
            for (String topic : topics) {
                config.saveChannelTopic(channel, topic);
                assertEquals(topic, config.getChannelTopic(channel));
            }
            assertEquals(config.getChannelTopic(channel), topics[topics.length - 1]);
        }

        for (String channel : channels) {
            assertEquals(config.getChannelTopic(channel), topics[topics.length - 1]);
        }
    }
}

